"""
Retrieve streaming scanner FFT results from a gr-iqtlabs/gamutRF scanner.

Example usage:

    from gamutrf.zmqreceiver import ZmqReceiver
    zmqr = ZmqReceiver(scanners=[("127.0.0.1", 8001)])

    while True:
        scan_configs, frame_df = zmqr.read_buffer()
        if frame_df is None:
            # no new scan result yet
            time.sleep(1)
        ...

    zmqr.stop()

scan_configs and frame_df will be None if the next full scan has not been received yet,
and scan updates are automatically processed in the background to avoid ZMQ tranport
overflows.

frame_df is a pandas DataFrame with the results of a full scan, and scan_configs is a
list of python dicts containing scan metadata - the contents of the "config" dict, from
https://github.com/IQTLabs/gr-iqtlabs/blob/main/grc/iqtlabs_retune_fft.block.yml.
"""

import concurrent.futures
import datetime
import tempfile
import json
import logging
import os
import pathlib
import re
import time
import zmq
import zstandard
import pandas as pd

FFT_BUFFER_TIME = 0.1
BUFF_FILE = "scanfftbuffer.txt.zst"  # nosec


def frame_resample(df, scan_fres):
    if df is not None:
        # ...first frequency
        df["freq"] = (df["freq"] / scan_fres).round() * scan_fres / 1e6
        df = df.set_index("freq")
        # ...then power
        df["db"] = df.groupby(["freq"])["db"].mean()
        df = df.reset_index().drop_duplicates(subset=["freq"])
        return df.sort_values("freq")
    return df


def parse_scanners(args_scanners):
    scanner_re = re.compile(r"^(.+):(\d+)$")
    scanners = []
    for scanner_str in args_scanners.split(","):
        scanner_match = scanner_re.match(scanner_str)
        if not scanner_match:
            raise ValueError(
                f"invalid scanner address: {scanner_str} from {args_scanners}"
            )
        scanners.append((scanner_match.group(1), int(scanner_match.group(2))))
    return scanners


def fft_proxy(
    addr, port, buff_file, buffer_time=FFT_BUFFER_TIME, live_file=None, poll_timeout=0.1
):
    zmq_addr = f"tcp://{addr}:{port}"
    logging.info("connecting to %s", zmq_addr)
    zmq_context = zmq.Context()
    socket = zmq_context.socket(zmq.SUB)
    socket.connect(zmq_addr)
    socket.setsockopt_string(zmq.SUBSCRIBE, "")
    packets_sent = 0
    last_packet_sent_time = time.time()
    tmp_buff_file = os.path.basename(buff_file)
    tmp_buff_file = buff_file.replace(tmp_buff_file, "." + tmp_buff_file)
    if os.path.exists(tmp_buff_file):
        os.remove(tmp_buff_file)
    context = zstandard.ZstdCompressor()
    shutdown = False
    while not shutdown:
        with open(tmp_buff_file, "wb") as zbf:
            with context.stream_writer(zbf) as bf:
                while not shutdown:
                    shutdown = live_file is not None and not live_file.exists()
                    try:
                        sock_txt = socket.recv(flags=zmq.NOBLOCK)
                    except zmq.error.Again:
                        time.sleep(poll_timeout)
                        continue
                    bf.write(sock_txt)
                    now = time.time()
                    if (
                        shutdown or now - last_packet_sent_time > buffer_time
                    ) and not os.path.exists(buff_file):
                        if packets_sent == 0:
                            logging.info("recording first FFT packet")
                        packets_sent += 1
                        last_packet_sent_time = now
                        break
        os.rename(tmp_buff_file, buff_file)


class ZmqScanner:
    def __init__(
        self,
        buff_path,
        proxy,
        addr,
        port,
        live_file,
        executor,
    ):
        self.buff_file = os.path.join(buff_path, "_".join((addr, str(port), BUFF_FILE)))
        if os.path.exists(self.buff_file):
            os.remove(self.buff_file)
        self.addr = addr
        self.port = port
        self.context = zstandard.ZstdDecompressor()
        self.txt_buf = ""
        self.fftbuffer = None
        self.scan_configs = {}
        self.proxy_result = executor.submit(
            proxy, addr, port, self.buff_file, live_file=live_file
        )

    def info(self, infostr):
        logging.info("%s:%u %s", self.addr, self.port, infostr)

    def healthy(self):
        return self.proxy_result.running()

    def __str__(self):
        return f"ZmqScanner on {self.addr}:{self.port}"

    def read_buff_file(self):
        if os.path.exists(self.buff_file):
            self.info("read %u bytes of FFT data" % os.stat(self.buff_file).st_size)
            with self.context.stream_reader(open(self.buff_file, "rb")) as bf:
                self.txt_buf += bf.read().decode("utf8")
            os.remove(self.buff_file)
            return True
        return False

    def txtbuf_to_lines(self, log):
        lines = self.txt_buf.splitlines()
        if len(lines) > 1:
            if self.txt_buf.endswith("\n"):
                if log:
                    log.write(self.txt_buf)
                self.txt_buf = ""
            elif lines:
                last_line = lines[-1]
                if log:
                    log.write(self.txt_buf[: -len(last_line)])
                self.txt_buf = last_line
                lines = lines[:-1]
            return lines
        return None

    def read_new_frame_df(self, df, discard_time):
        frame_df = None
        scan_config = None
        if discard_time:
            df = df[(time.time() - df.ts).abs() < discard_time]
        if df.size:
            lastfreq = df["freq"].iat[-1]
            self.info("last frequency read %f MHz" % (lastfreq / 1e6))
            if self.fftbuffer is None:
                self.fftbuffer = df
            else:
                self.fftbuffer = pd.concat([self.fftbuffer, df])
            if self.fftbuffer["sweep_start"].nunique() > 1:
                min_sweep_start = self.fftbuffer["sweep_start"].min()
                frame_df = self.fftbuffer[
                    self.fftbuffer["sweep_start"] == min_sweep_start
                ].copy()
                frame_df["tune_count"] = (
                    frame_df["tune_count"].max() - frame_df["tune_count"].min()
                )
                self.fftbuffer = self.fftbuffer[
                    self.fftbuffer["sweep_start"] != min_sweep_start
                ]
                scan_config = self.scan_configs[min_sweep_start]
                del self.scan_configs[min_sweep_start]
        return (scan_config, frame_df)

    def lines_to_df(self, lines):
        try:
            records = []
            for line in lines:
                line = line.strip()
                json_record = json.loads(line)
                ts = float(json_record["ts"])
                sweep_start = float(json_record["sweep_start"])
                total_tune_count = int(json_record["total_tune_count"])
                buckets = json_record["buckets"]
                scan_config = json_record["config"]
                self.scan_configs[sweep_start] = scan_config
                records.extend(
                    [
                        {
                            "ts": ts,
                            "freq": float(freq),
                            "db": float(db),
                            "sweep_start": sweep_start,
                            "tune_count": total_tune_count,
                        }
                        for freq, db in buckets.items()
                    ]
                )
            return pd.DataFrame(records)
        except ValueError as err:
            logging.error(str(err))
            return None

    def read_buff(self, log, discard_time):
        scan_config = None
        frame_df = None
        if self.read_buff_file():
            lines = self.txtbuf_to_lines(log)
            if lines:
                df = self.lines_to_df(lines)
                scan_config, frame_df = self.read_new_frame_df(df, discard_time)
        return scan_config, frame_df


class ZmqReceiver:
    def __init__(
        self,
        scanners=[("127.0.0.1", 8001)],
        buff_path=None,
        proxy=fft_proxy,
    ):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.live_file = pathlib.Path(os.path.join(self.tmpdir.name, "live_file"))
        self.live_file.touch()
        if buff_path is None:
            buff_path = self.tmpdir.name
        self.executor = concurrent.futures.ProcessPoolExecutor(len(scanners))
        self.scanners = []
        self.last_results = []
        for addr, port in scanners:
            self.scanners.append(
                ZmqScanner(buff_path, proxy, addr, port, self.live_file, self.executor)
            )

    def stop(self):
        self.live_file.unlink()
        self.executor.shutdown()
        self.tmpdir.cleanup()

    def healthy(self):
        if os.path.exists(self.live_file):
            for scanner in self.scanners:
                if not scanner.healthy():
                    return False
            return True
        return False

    def read_buff(self, log=None, discard_time=0, scan_fres=0):
        results = [scanner.read_buff(log, discard_time) for scanner in self.scanners]
        if self.last_results:
            for i, result in enumerate(results):
                _scan_config, df = result
                if df is not None:
                    self.last_results[i] = result
                    logging.info("%s got scan result", self.scanners[i])
        else:
            self.last_results = results

        scan_configs = []
        dfs = []

        for scan_config, df in self.last_results:
            scan_configs.append(scan_config)
            if df is not None:
                dfs.append(df)

        df = None
        if len(dfs) == len(self.scanners):
            df = pd.concat(dfs)
            logging.info(
                "all scanners got result, %s to %s",
                datetime.datetime.fromtimestamp(df.ts.min()).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                datetime.datetime.fromtimestamp(df.ts.max()).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
            )
            if scan_fres:
                df = frame_resample(df, scan_fres)
            self.last_results = []

        return (scan_configs, df)



import json
import logging
import os
import socket

import gpsd
import httpx
import paho.mqtt.client as mqtt


class MQTTReporter:
    def __init__(self, name, mqtt_server=None, gps_server=None, compass=False):
        self.name = name
        self.mqtt_server = mqtt_server
        self.compass = compass
        self.gps_server = gps_server
        self.mqttc = None
        self.heading = "no heading"

    @staticmethod
    def log(path, prefix, start_time, record_args):
        try:
            with open(
                os.path.join(path, f"mqtt-{prefix}-{start_time}.log"),
                "a+",
                encoding="utf-8",
            ) as f:
                f.write(f"{json.dumps(record_args)}\n")
        except FileNotFoundError as err:
            logging.error(f"could not write to mqtt rssi log: {err}")

    def connect(self):
        logging.info(f"connecting to {self.mqtt_server}")
        self.mqttc = mqtt.Client()
        self.mqttc.connect(self.mqtt_server)
        self.mqttc.loop_start()

    def get_heading(self):
        try:
            self.heading = str(
                float(httpx.get(f"http://{self.gps_server}:8000/v1/heading").text)
            )
        except Exception as err:
            logging.error("could not update heading: %s", err)

    def add_gps(self, publish_args):
        if not self.gps_server:
            return publish_args
        publish_args.update(
            {
                "position": [0, 0],
                "altitude": None,
                "gps_time": None,
                "map_url": None,
                "heading": self.heading,
                "gps": "no fix",
            }
        )
        try:
            if self.compass:
                self.get_heading()
            if gpsd.gpsd_stream is None:
                gpsd.connect(host=self.gps_server, port=2947)
            packet = gpsd.get_current()
            publish_args.update(
                {
                    "position": packet.position(),
                    "altitude": packet.altitude(),
                    "gps_time": packet.get_time().timestamp(),
                    "map_url": packet.map_url(),
                    "heading": self.heading,
                    "gps": "fix",
                }
            )
        except (BrokenPipeError, gpsd.NoFixError, AttributeError) as err:
            logging.error("could not update with GPS: %s", err)
        return publish_args

    def publish(self, publish_path, publish_args):
        if not self.mqtt_server:
            return
        try:
            if self.mqttc is None:
                self.connect()
            publish_args = self.add_gps(publish_args)
            publish_args["name"] = self.name
            self.mqttc.publish(publish_path, json.dumps(publish_args))
        except (
            socket.gaierror,
            ConnectionRefusedError,
            mqtt.WebsocketConnectionError,
            ValueError,
        ) as err:
            logging.error(f"failed to publish to MQTT {self.mqtt_server}: {err}")



#!/usr/bin/env python3

import gzip
import zstandard
import numpy as np


def get_reader(filename):
    # nosemgrep:github.workflows.config.useless-inner-function
    def gzip_reader(x):
        return gzip.open(x, "rb")

    # nosemgrep:github.workflows.config.useless-inner-function
    def zst_reader(x):
        return zstandard.ZstdDecompressor().stream_reader(
            open(x, "rb"), read_across_frames=True
        )

    def default_reader(x):
        return open(x, "rb")

    if filename.endswith(".gz"):
        return gzip_reader
    if filename.endswith(".zst"):
        return zst_reader

    return default_reader


def read_recording(
    filename,
    sample_rate,
    sample_dtype,
    sample_len,
    sample_secs=1.0,
    skip_sample_secs=0,
    max_sample_secs=0,
):
    """Read an I/Q recording and iterate over it, returning 1-D numpy arrays of csingles, of size sample_rate * sample_secs.

    Args:
        filename: str, recording to read.
        sample_rate: int, samples per second
        sample_dtype: numpy.dtype, binary format of original I/Q recording.
        sample_len: int, length of one sample.
        sample_secs: float, number of seconds worth of samples per iteration.
        skip_sample_secs: float, number of seconds worth of samples to skip initially.
        max_sample_secs: float, maximum number of seconds of samples to read (or None for all).
    Returns:
        numpy arrays of csingles.
    """
    read_size = int(sample_rate * sample_secs) * sample_len
    reader = get_reader(filename)
    samples_remaining = 0
    if max_sample_secs:
        samples_remaining = int(sample_rate * max_sample_secs)
    with reader(filename) as infile:
        if skip_sample_secs:
            infile.seek(int(sample_rate * skip_sample_secs) * sample_len)
        while True:
            if max_sample_secs and not samples_remaining:
                break
            sample_buffer = infile.read(read_size)
            buffered_samples = int(len(sample_buffer) / sample_len)
            if buffered_samples == 0:
                break
            if max_sample_secs:
                if buffered_samples <= samples_remaining:
                    samples_remaining -= buffered_samples
                else:
                    buffered_samples = samples_remaining
                    samples_remaining = 0
            x1d = np.frombuffer(
                sample_buffer, dtype=sample_dtype, count=buffered_samples
            )
            yield x1d["i"] + np.csingle(1j) * x1d["q"]



import argparse
import subprocess

from gamutrf.utils import parse_filename
from gamutrf.utils import replace_ext


def make_procs_args(sample_filename, outfmt):
    procs_args = []
    out_filename = sample_filename

    if sample_filename.endswith(".gz"):
        procs_args.append(["gunzip", "-c", sample_filename])
        out_filename = replace_ext(out_filename, "")
    elif sample_filename.endswith(".zst"):
        procs_args.append(["zstdcat", sample_filename])
        out_filename = replace_ext(out_filename, "")

    _, _, sample_rate, _sample_dtype, _sample_len, in_format, in_bits = parse_filename(
        out_filename
    )
    print(_, sample_rate, _sample_dtype, _sample_len, in_format, in_bits)
    out_filename = replace_ext(out_filename, "raw", all_ext=True)
    sox_in = sample_filename
    if procs_args:
        sox_in = "-"
    procs_args.append(
        [
            "sox",
            "-D",  # disable dithering.
            "-t",
            "raw",
            "-r",
            str(sample_rate),
            "-c",
            "1",
            "-b",
            str(in_bits),
            "-e",
            in_format,
            sox_in,
            "-e",
            outfmt,
            out_filename,
        ]
    )
    return procs_args


def run_procs(procs_args):
    procs = []
    for proc_args in procs_args:
        stdin = None
        if procs:
            stdin = procs[-1].stdout
        procs.append(subprocess.Popen(proc_args, stdout=subprocess.PIPE, stdin=stdin))
    for proc in procs[:-1]:
        if proc.stdout is not None:
            proc.stdout.close()
    procs[-1].communicate()


def argument_parser():
    parser = argparse.ArgumentParser(
        description="Convert (possibly compressed) sample recording to a float raw file (gnuradio style)"
    )
    parser.add_argument("samplefile", default="", help="sample file to read")
    parser.add_argument("--outfmt", default="float", help="output format")
    return parser


def main():
    parser = argument_parser()
    args = parser.parse_args()
    procs_args = make_procs_args(args.samplefile, args.outfmt)
    run_procs(procs_args)



# Extract and decimate narrowband signal from wideband signal
#
# Example of decoding FM signal centered at 98.1MHz, from 20MHz wideband recording centered at 100MHz recorded at 20Msps.
#
# Convert int16 recording to complex float:
#
# $ samples2raw.py gamutrf_recording1643863029_100000000Hz_20000000sps.s16.gz
#
# Select new center -1.9MHz, from original center and downsample to 1Msps
#
# $ freqxlator.py --samp-rate=20e6 --center=-1.9e6 --dec=20 --infile gamutrf_recording1643863029_100000000Hz_20000000sps.raw --outfile fm.raw
#
# Decode 1Msps recording as FM.
#
# $ airspy-fmradion -t filesource -c srate=1000000,raw,format=FLOAT,filename=fm.raw -W fm.wav
import argparse
import sys

try:
    from gnuradio import blocks  # pytype: disable=import-error
    from gnuradio import eng_notation  # pytype: disable=import-error
    from gnuradio import gr  # pytype: disable=import-error
    from gnuradio.eng_arg import eng_float  # pytype: disable=import-error
    from gnuradio.filter import firdes
    from gnuradio.filter import freq_xlating_fir_filter_ccc
    from pmt import PMT_NIL  # pytype: disable=import-error
except ModuleNotFoundError as err:
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)


class FreqXLator(gr.top_block):
    def __init__(self, samp_rate, center, transitionbw, dec, infile, outfile):
        gr.top_block.__init__(self, "freqxlator", catch_exceptions=True)

        self.samp_rate = samp_rate
        self.center = center
        self.transitionbw = transitionbw
        self.dec = dec
        self.infile = infile
        self.outfile = outfile

        self.freq_xlating_fir_filter_xxx_0 = freq_xlating_fir_filter_ccc(
            self.dec, self._get_taps(), self.center, self.samp_rate
        )
        self.blocks_file_source_0 = blocks.file_source(
            gr.sizeof_gr_complex, self.infile, False, 0, 0
        )  # pylint: disable=no-member
        self.blocks_file_source_0.set_begin_tag(PMT_NIL)
        self.blocks_file_sink_0 = blocks.file_sink(
            gr.sizeof_gr_complex, self.outfile, False
        )  # pylint: disable=no-member
        self.blocks_file_sink_0.set_unbuffered(False)

        self.connect(
            (self.blocks_file_source_0, 0), (self.freq_xlating_fir_filter_xxx_0, 0)
        )
        self.connect(
            (self.freq_xlating_fir_filter_xxx_0, 0), (self.blocks_file_sink_0, 0)
        )

    def _get_taps(self):
        return firdes.complex_band_pass(
            1,
            self.samp_rate,
            -self.samp_rate / (2 * self.dec),
            self.samp_rate / (2 * self.dec),
            self.transitionbw,
        )


def argument_parser():
    parser = argparse.ArgumentParser(
        "Extract and decimate narrowband signal from wideband signal"
    )
    parser.add_argument("infile", type=str, help="Input file (complex I/Q format)")
    parser.add_argument("outfile", type=str, help="Output file (complex I/Q format)")
    parser.add_argument(
        "--samp-rate",
        dest="samp_rate",
        type=eng_float,
        default=eng_notation.num_to_str(float(20e6)),
        help="Set samp_rate [default=%(default)r]",
    )
    parser.add_argument(
        "--center",
        dest="center",
        type=eng_float,
        default=eng_notation.num_to_str(float(-1e6)),
        help="Offset to new center frequency [default=%(default)r]",
    )
    parser.add_argument(
        "--transitionbw",
        dest="transitionbw",
        type=eng_float,
        default=eng_notation.num_to_str(float(10e3)),
        help="Filter transmission bandwidth [default=%(default)r]",
    )
    parser.add_argument(
        "--dec",
        dest="dec",
        type=int,
        default=20,
        help="Decimation [default=%(default)r]",
    )
    return parser


def main():
    parser = argument_parser()
    options = parser.parse_args()
    block = FreqXLator(
        options.samp_rate,
        options.center,
        options.transitionbw,
        options.dec,
        options.infile,
        options.outfile,
    )
    block.start()
    block.wait()



import logging
import os
import sys
import time
from urllib.parse import urlparse

try:
    import pmt
    from gnuradio import blocks
    from gnuradio import soapy
    from gnuradio import uhd
    from gnuradio.gr import sizeof_gr_complex
    from gnuradio import iqtlabs
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

from gamutrf.ettus_source import get_ettus_source
from gamutrf.soapy_source import get_soapy_source


def null_workaround_start_hook(self):
    return


def airt_workaround_start_hook(self):
    logging.info("applying AIRT workarounds")
    workaround_time = 0.5
    for rate in (15.625e6, self.samp_rate):
        time.sleep(workaround_time)
        self.sources[0].set_sample_rate(0, rate)
        self.sources[0].set_bandwidth(0, rate)


def get_source(
    sdr,
    samp_rate,
    gain,
    agc=False,
    center_freq=None,
    sdrargs=None,
    soapy_lib=soapy,
    uhd_lib=uhd,
):
    logging.info(
        f"initializing SDR {sdr} with sample rate {samp_rate}, gain {gain}, agc {agc}"
    )

    workaround_start_hook = null_workaround_start_hook
    cmd_port = "command"
    sources = []

    url = urlparse(sdr)
    if url.scheme:
        if url.scheme == "file" and os.path.exists(url.path):
            sources = [
                blocks.file_source(sizeof_gr_complex, url.path, True, 0, 0),
                blocks.throttle(sizeof_gr_complex, samp_rate, True),
            ]
            sources[0].message_port_register_in(pmt.intern(cmd_port))
        else:
            raise ValueError("unsupported/missing file location")
    elif sdr == "tuneable_test_source":
        freq_divisor = 1e9
        cmd_port = "cmd"
        sources = [
            iqtlabs.tuneable_test_source(freq_divisor),
            blocks.throttle(sizeof_gr_complex, samp_rate, True),
        ]
    elif sdr == "ettus":
        sources = get_ettus_source(sdrargs, samp_rate, center_freq, agc, gain, uhd_lib)
    else:
        sources, cmd_port = get_soapy_source(
            sdr, sdrargs, samp_rate, center_freq, agc, gain, soapy_lib
        )
        if sdr == "SoapyAIRT":
            workaround_start_hook = airt_workaround_start_hook

    return sources, cmd_port, workaround_start_hook



"""Main entrypoint for GamutRF"""
from gamutrf.api import main as api_main
from gamutrf.compress_dirs import main as compress_dirs_main
from gamutrf.freqxlator import main as freqxlator_main
from gamutrf.samples2raw import main as samples2raw_main
from gamutrf.scan import main as scan_main
from gamutrf.sigfinder import main as sigfinder_main
from gamutrf.specgram import main as specgram_main
from gamutrf.waterfall import main as waterfall_main


def api():
    """Entrypoint for API"""
    api_main()


def compress_dirs():
    """Entrypoint for compress_dirs"""
    compress_dirs_main()


def freqxlator():
    """Entrypoint for freqxlator"""
    freqxlator_main()


def samples2raw():
    """Entrypoint for samples2raw"""
    samples2raw_main()


def scan():
    """Entrypoint for scan"""
    scan_main()


def sigfinder():
    """Entrypoint for sigfinder"""
    sigfinder_main()


def specgram():
    """Entrypoint for specgram"""
    specgram_main()


def waterfall():
    """Entrypoint for waterfall"""
    waterfall_main()



import argparse
import csv
import datetime
import json
import logging
import multiprocessing
import os
import shutil
import signal
import tempfile
import time
import warnings
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from flask import Flask, current_app
from matplotlib.collections import LineCollection
from matplotlib.ticker import MultipleLocator, AutoMinorLocator
from matplotlib import style
from scipy.ndimage import gaussian_filter

from gamutrf.peak_finder import get_peak_finder
from gamutrf.utils import SCAN_FRES
from gamutrf.zmqreceiver import ZmqReceiver, parse_scanners, frame_resample

warnings.filterwarnings(action="ignore", message="Mean of empty slice")
warnings.filterwarnings(action="ignore", message="All-NaN slice encountered")
warnings.filterwarnings(action="ignore", message="Degrees of freedom <= 0 for slice.")


def safe_savefig(path):
    basename = os.path.basename(path)
    dirname = os.path.dirname(path)
    tmp_path = os.path.join(dirname, "." + basename)
    plt.savefig(tmp_path)
    os.rename(tmp_path, path)
    logging.info("wrote %s", path)
    return path


def draw_title(
    ax,
    title,
    scan_duration,
    tune_step_hz,
    tune_step_fft,
    tune_rate_hz,
    tune_dwell_ms,
    sample_rate,
    freq_resolution,
):
    title_text = {
        "Time": str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
        "Scan time": "%.2fs" % scan_duration,
        "Step FFTs": "%u" % tune_step_fft,
        "Step size": "%.2fMHz" % (tune_step_hz / 1e6),
        "Sample rate": "%.2fMsps" % (sample_rate / 1e6),
        "Resolution": "%.2fMHz" % freq_resolution,
        "Tune rate": "%.2fHz" % tune_rate_hz,
        "Tune dwell time": "%.2fms" % tune_dwell_ms,
    }
    title.set_fontsize(8)
    title.set_text(str(title_text))
    ax.draw_artist(title)


def filter_peaks(peaks, properties):
    for i in range(len(peaks) - 1, -1, -1):  # start from end of list
        for j in range(len(peaks)):
            if i == j:
                continue
            if (properties["left_ips"][i] > properties["left_ips"][j]) and (
                properties["right_ips"][i] < properties["right_ips"][j]
            ):
                peaks = np.delete(peaks, i)
                for k in properties:
                    properties[k] = np.delete(properties[k], i)

                break
                # properties["left_ips"] = np.delete(properties["left_ips"], i)
                # properties["right_ips"] = np.delete(properties["right_ips"], i)
                # properties["width_heights"] = np.delete(properties["width_heights"], i)
    return peaks, properties


def save_detections(
    config,
    state,
    scan_time,
    scan_configs,
    peaks,
    properties,
):
    detection_save_dir = Path(state.save_path, "detections")

    detection_config_save_path = str(
        Path(
            detection_save_dir,
            f"detections_scan_config_{scan_time}.json",
        )
    )
    detection_save_path = str(
        Path(
            detection_save_dir,
            f"detections_{scan_time}.csv",
        )
    )

    if not os.path.exists(detection_save_dir):
        Path(detection_save_dir).mkdir(parents=True, exist_ok=True)

    if state.previous_scan_config is None or state.previous_scan_config != scan_configs:
        state.previous_scan_config = scan_configs
        with open(detection_config_save_path, "w", encoding="utf8") as f:
            json.dump(
                {
                    "timestamp": scan_time,
                    "min_freq": config.min_freq,
                    "max_freq": config.max_freq,
                    "scan_configs": scan_configs,
                },
                f,
                indent=4,
            )

    if not os.path.exists(detection_save_path):
        with open(detection_save_path, "w", encoding="utf8") as detection_csv:
            writer = csv.writer(detection_csv)
            writer.writerow(
                [
                    "timestamp",
                    "start_freq",
                    "end_freq",
                    "dB",
                    "type",
                ]
            )

    with open(detection_save_path, "a", encoding="utf8") as detection_csv:
        writer = csv.writer(detection_csv)
        for i in range(len(peaks)):
            writer.writerow(
                [
                    scan_time,  # timestamp
                    state.psd_x_edges[
                        properties["left_ips"][i].astype(int)
                    ],  # start_freq
                    state.psd_x_edges[
                        properties["right_ips"][i].astype(int)
                    ],  # end_freq
                    properties["peak_heights"][i],  # dB
                    state.peak_finder.name,  # type
                ]
            )


def save_waterfall(
    state,
    save_time,
    scan_time,
    fig_path=None,
):
    now = datetime.datetime.now()
    if state.last_save_time is None:
        state.last_save_time = now

    if now - state.last_save_time > datetime.timedelta(minutes=save_time):
        waterfall_save_dir = Path(state.save_path, "waterfall")
        if not os.path.exists(waterfall_save_dir):
            Path(waterfall_save_dir).mkdir(parents=True, exist_ok=True)

        waterfall_save_path = str(
            Path(waterfall_save_dir, f"waterfall_{scan_time}.png")
        )
        if fig_path:
            shutil.copyfile(fig_path, waterfall_save_path)
        else:
            safe_savefig(waterfall_save_path)

        save_scan_configs = {
            "start_scan_timestamp": state.scan_times[0],
            "start_scan_config": state.scan_config_history[state.scan_times[0]],
            "end_scan_timestamp": state.scan_times[-1],
            "end_scan_config": state.scan_config_history[state.scan_times[-1]],
        }
        config_save_path = str(Path(waterfall_save_dir, f"config_{scan_time}.json"))
        with open(config_save_path, "w", encoding="utf8") as f:
            json.dump(save_scan_configs, f, indent=4)

        state.last_save_time = now
        logging.info(f"Saving {waterfall_save_path}")


def argument_parser():
    parser = argparse.ArgumentParser(description="waterfall plotter from scan data")
    parser.add_argument(
        "--min_freq",
        default=0,
        type=float,
        help="Minimum frequency for plot (or 0 for automatic).",
    )
    parser.add_argument(
        "--max_freq",
        default=0,
        type=float,
        help="Maximum frequency for plot (or 0 for automatic).",
    )
    parser.add_argument(
        "--n_detect", default=0, type=int, help="Number of detected signals to plot."
    )
    parser.add_argument(
        "--plot_snr", action="store_true", help="Plot SNR rather than power."
    )
    parser.add_argument(
        "--detection_type",
        default="",
        type=str,
        help="Detection type to plot (wideband, narrowband).",
    )
    parser.add_argument(
        "--save_path", default="", type=str, help="Path to save screenshots."
    )
    parser.add_argument(
        "--save_time",
        default=1,
        type=int,
        help="Save screenshot every save_time minutes. Only used if save_path also defined.",
    )
    parser.add_argument(
        "--scanners",
        default="127.0.0.1:8001",
        type=str,
        help="Scanner endpoints to use.",
    )
    parser.add_argument(
        "--port",
        default=0,
        type=int,
        help="If set, serve waterfall on this port.",
    )
    parser.add_argument(
        "--rotate_secs",
        default=900,
        type=int,
        help="If > 0, rotate save directories every N seconds",
    )
    parser.add_argument(
        "--width",
        default=28,
        type=float,
        help="Waterfall image width",
    )
    parser.add_argument(
        "--height",
        default=10,
        type=float,
        help="Waterfall image height",
    )
    parser.add_argument(
        "--waterfall_height",
        default=100,
        type=int,
        help="Waterfall height",
    )
    parser.add_argument(
        "--waterfall_width",
        default=5000,
        type=int,
        help="Waterfall width (maximum)",
    )

    return parser


def reset_mesh_psd(config, state, data=None):
    if state.mesh_psd:
        state.mesh_psd.remove()

    X, Y = meshgrid(config, state.db_min, state.db_max, config.psd_db_resolution)
    state.psd_x_edges = X[0]
    state.psd_y_edges = Y[:, 0]

    if data is None:
        data = np.zeros(X[:-1, :-1].shape)

    state.mesh_psd = state.ax_psd.pcolormesh(X, Y, data, shading="flat")


def reset_mesh(state, data):
    if state.mesh:
        state.mesh.remove()
    state.mesh = state.ax.pcolormesh(state.X, state.Y, data, shading="nearest")


def reset_fig(
    config,
    state,
):
    logging.info("resetting figure")

    state.fig.clf()
    plt.tight_layout()
    plt.subplots_adjust(hspace=0.15)
    state.ax_psd = state.fig.add_subplot(3, 1, 1)
    state.ax = state.fig.add_subplot(3, 1, (2, 3))
    state.psd_title = state.ax_psd.text(
        0.5,
        1.05,
        "",
        transform=state.ax_psd.transAxes,
        va="center",
        ha="center",
    )
    default_data = state.db_min * np.ones(state.freq_data.shape[1])

    reset_mesh_psd(config, state)

    def ax_psd_plot(linestyle=":", **kwargs):
        return state.ax_psd.plot(
            state.X[0],
            default_data,
            markevery=config.marker_distance,
            linestyle=linestyle,
            **kwargs,
        )

    (state.peak_lns,) = ax_psd_plot(
        color="white",
        marker="^",
        markersize=12,
        linestyle="none",
        fillstyle="full",
    )
    (state.max_psd_ln,) = ax_psd_plot(
        color="red",
        marker=",",
        label="max",
    )
    (state.min_psd_ln,) = ax_psd_plot(
        color="pink",
        marker=",",
        label="min",
    )
    (state.mean_psd_ln,) = ax_psd_plot(
        color="cyan",
        marker="^",
        markersize=8,
        fillstyle="none",
        label="mean",
    )
    (state.current_psd_ln,) = ax_psd_plot(
        color="red",
        marker="o",
        markersize=8,
        fillstyle="none",
        label="current",
    )
    state.ax_psd.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    state.ax_psd.set_ylabel("dB")

    # SPECTROGRAM
    reset_mesh(state, state.db_data)
    state.top_n_lns = []
    for _ in range(config.top_n):
        (ln,) = state.ax.plot(
            [state.X[0][0]] * len(state.Y[:, 0]),
            state.Y[:, 0],
            color="brown",
            linestyle=":",
            alpha=0,
        )
        ln.set_alpha(0.75)
        state.top_n_lns.append(ln)

    state.ax.set_xlabel("MHz")
    state.ax.set_ylabel("Time")

    # COLORBAR
    state.sm = plt.cm.ScalarMappable(cmap=state.cmap)
    state.sm.set_clim(vmin=state.db_min, vmax=state.db_max)

    if config.plot_snr:
        state.sm.set_clim(vmin=config.snr_min, vmax=config.snr_max)
    state.cbar_ax = state.fig.add_axes([0.92, 0.10, 0.03, 0.5])
    state.cbar = state.fig.colorbar(state.sm, cax=state.cbar_ax)
    state.cbar.set_label("dB", rotation=0)

    # SPECTROGRAM TITLE
    _title = state.ax.text(
        0.5, 1.05, "", transform=state.ax.transAxes, va="center", ha="center"
    )

    for ax in (state.ax.xaxis, state.ax_psd.xaxis):
        ax.set_major_locator(state.major_tick_separator)
        if config.freq_resolution < 0.01:
            ax.set_major_formatter("{x:.1f}")
        else:
            ax.set_major_formatter("{x:.0f}")
        ax.set_minor_locator(state.minor_tick_separator)

    for ax in (state.ax_psd.yaxis, state.cbar_ax.yaxis, state.ax.yaxis):
        ax.set_animated(True)

    state.ax.draw_artist(state.mesh)
    state.fig.canvas.blit(state.ax.bbox)

    if not config.batch:
        plt.show(block=False)
        state.fig.canvas.flush_events()
        state.background = state.fig.canvas.copy_from_bbox(state.fig.bbox)
        if config.savefig_path:
            safe_savefig(config.savefig_path)


def meshgrid(config, start, stop, num):
    return np.meshgrid(
        np.linspace(
            config.min_freq,
            config.max_freq,
            config.waterfall_width,
        ),
        np.linspace(start, stop, num),
    )


def init_fig(
    config,
    state,
    onresize,
):
    logging.info("initializing figure")
    plt.close("all")

    state.cmap = plt.get_cmap("viridis")
    state.cmap_psd = plt.get_cmap("turbo")
    state.minor_tick_separator = AutoMinorLocator()
    state.major_tick_separator = MultipleLocator(config.freq_range / config.n_ticks)

    plt.rcParams["savefig.facecolor"] = "#2A3459"
    plt.rcParams["figure.facecolor"] = "#2A3459"
    for param in (
        "text.color",
        "axes.labelcolor",
        "xtick.color",
        "ytick.color",
        "axes.facecolor",
    ):
        plt.rcParams[param] = "#d2d5dd"

    state.fig = plt.figure(figsize=(config.width, config.height), dpi=100)
    if not config.batch:
        state.fig.canvas.mpl_connect("resize_event", onresize)


def draw_peaks(
    config,
    state,
    scan_time,
    scan_configs,
):
    peaks, properties = state.peak_finder.find_peaks(state.db_data[-1])
    peaks, properties = filter_peaks(peaks, properties)
    left_ips = properties["left_ips"].astype(int)
    right_ips = properties["right_ips"].astype(int)

    if state.save_path:
        save_detections(
            config,
            state,
            scan_time,
            scan_configs,
            peaks,
            properties,
        )

    state.peak_lns.set_xdata(state.psd_x_edges[peaks])
    state.peak_lns.set_ydata(properties["width_heights"])

    for child in state.ax_psd.get_children():
        if isinstance(child, LineCollection):
            child.remove()

    for i in range(len(state.detection_text) - 1, -1, -1):
        state.detection_text[i].set_visible(False)
        state.detection_text.pop(i)

    if len(peaks) > 0:
        vl_center = state.ax_psd.vlines(
            x=state.psd_x_edges[peaks],
            ymin=state.db_data[-1][peaks] - properties["prominences"],
            ymax=state.db_data[-1][peaks],
            color="white",
        )
        state.ax_psd.draw_artist(vl_center)
        vl_edges = state.ax_psd.vlines(
            x=np.concatenate(
                (
                    state.psd_x_edges[left_ips],
                    state.psd_x_edges[right_ips],
                )
            ),
            ymin=state.db_min,
            ymax=np.tile(state.db_data[-1][peaks], 2),
            color="white",
        )
        state.ax_psd.draw_artist(vl_edges)
        for l_ips, r_ips, p in zip(
            state.psd_x_edges[left_ips],
            state.psd_x_edges[right_ips],
            state.db_data[-1][peaks],
        ):
            shaded = state.ax_psd.fill_between(
                [l_ips, r_ips], state.db_min, p, alpha=0.7
            )
            state.ax_psd.draw_artist(shaded)
        hl = state.ax_psd.hlines(
            y=properties["width_heights"],
            xmin=state.psd_x_edges[left_ips],
            xmax=state.psd_x_edges[right_ips],
            color="white",
        )
        state.ax_psd.draw_artist(hl)
        for l_ips, r_ips, p in zip(
            state.psd_x_edges[left_ips],
            state.psd_x_edges[right_ips],
            peaks,
        ):
            for txt in (
                state.ax_psd.text(
                    l_ips + ((r_ips - l_ips) / 2),
                    (0.15 * (state.db_max - state.db_min)) + state.db_min,
                    f"f={l_ips + ((r_ips - l_ips)/2):.0f}MHz",
                    size=10,
                    ha="center",
                    color="white",
                    rotation=40,
                ),
                state.ax_psd.text(
                    l_ips + ((r_ips - l_ips) / 2),
                    (0.05 * (state.db_max - state.db_min)) + state.db_min,
                    f"BW={r_ips - l_ips:.0f}MHz",
                    size=10,
                    ha="center",
                    color="white",
                    rotation=40,
                ),
            ):
                state.detection_text.append(txt)
                state.ax_psd.draw_artist(txt)


def update_fig(config, state, results):
    if not state.fig or not state.ax:
        raise NotImplementedError

    if config.base_save_path and config.rotate_secs:
        state.save_path = os.path.join(
            config.base_save_path,
            str(int(time.time() / config.rotate_secs) * config.rotate_secs),
        )
        if not os.path.exists(state.save_path):
            Path(state.save_path).mkdir(parents=True, exist_ok=True)

    if len(results) > 1:
        logging.info("processing backlog of %u results", len(results))

    scan_duration = 0

    for scan_configs, scan_df in results:
        tune_step_hz = min(scan_config["tune_step_hz"] for scan_config in scan_configs)
        tune_step_fft = min(
            scan_config["tune_step_fft"] for scan_config in scan_configs
        )
        scan_duration = scan_df.ts.max() - scan_df.ts.min()
        tune_count = scan_df.tune_count.max()
        tune_rate_hz = tune_count / scan_duration
        tune_dwell_ms = (scan_duration * 1e3) / tune_count
        idx = (
            ((scan_df.freq - config.min_freq) / config.freq_resolution)
            .round()
            .clip(lower=0, upper=(state.db_data.shape[1] - 1))
            .values.flatten()
            .astype(int)
        )

        state.freq_data = np.roll(state.freq_data, -1, axis=0)
        state.freq_data[-1][idx] = scan_df.freq.values.flatten()

        state.db_data = np.roll(state.db_data, -1, axis=0)
        state.db_data[-1][idx] = scan_df.db.values.flatten()

        scan_time = scan_df.ts.iloc[-1]
        row_time = datetime.datetime.fromtimestamp(scan_time)
        state.scan_times.append(scan_time)
        state.scan_config_history[scan_time] = scan_configs
        while len(state.scan_times) > config.waterfall_height:
            remove_time = state.scan_times.pop(0)
            state.scan_config_history.pop(remove_time)

        if state.counter % config.y_label_skip == 0:
            state.y_labels.append(row_time.strftime("%Y-%m-%d %H:%M:%S"))
        else:
            state.y_labels.append("")
        state.y_ticks.append(config.waterfall_height)
        for j in range(len(state.y_ticks) - 2, -1, -1):
            state.y_ticks[j] -= 1
            if state.y_ticks[j] < 1:
                state.y_ticks.pop(j)
                state.y_labels.pop(j)

        state.counter += 1

    if state.counter % config.draw_rate == 0:
        now = time.time()
        since_last_plot = 0
        if state.last_plot:
            since_last_plot = now - state.last_plot
        state.last_plot = now
        logging.info(f"Plotting {row_time} (seconds since last plot {since_last_plot})")

        state.db_min = np.nanmin(state.db_data)
        state.db_max = np.nanmax(state.db_data)

        data, _xedge, _yedge = np.histogram2d(
            state.freq_data[~np.isnan(state.freq_data)].flatten(),
            state.db_data[~np.isnan(state.db_data)].flatten(),
            density=False,
            bins=[state.psd_x_edges, state.psd_y_edges],
        )
        heatmap = gaussian_filter(data, sigma=2)
        data = heatmap / np.max(heatmap)

        db_norm = (state.db_data - state.db_min) / (state.db_max - state.db_min)
        if config.plot_snr:
            db_norm = (
                (state.db_data - np.nanmin(state.db_data, axis=0)) - config.snr_min
            ) / (config.snr_max - config.snr_min)

        top_n_bins = state.freq_bins[
            np.argsort(
                np.nanvar(state.db_data - np.nanmin(state.db_data, axis=0), axis=0)
            )[::-1][: config.top_n]
        ]

        state.ax.set_yticks(state.y_ticks, labels=state.y_labels)

        if state.background:
            state.fig.canvas.restore_region(state.background)

        for top_n_bin, ln in zip(top_n_bins, state.top_n_lns):
            ln.set_xdata([top_n_bin] * len(state.Y[:, 0]))

        state.fig.canvas.blit(state.ax.yaxis.axes.figure.bbox)

        reset_mesh_psd(config, state, data=state.cmap_psd(data.T))

        state.ax_psd.set_ylim(state.db_min, state.db_max)
        state.current_psd_ln.set_ydata(state.db_data[-1])
        for ln, ln_func in (
            (state.min_psd_ln, np.nanmin),
            (state.max_psd_ln, np.nanmax),
            (state.mean_psd_ln, np.nanmean),
        ):
            ln.set_ydata(ln_func(state.db_data, axis=0))
        state.ax_psd.draw_artist(state.mesh_psd)

        if state.peak_finder:
            draw_peaks(
                config,
                state,
                scan_time,
                scan_configs,
            )

        for ln in (
            state.peak_lns,
            state.min_psd_ln,
            state.max_psd_ln,
            state.mean_psd_ln,
            state.current_psd_ln,
        ):
            state.ax_psd.draw_artist(ln)

        reset_mesh(state, state.cmap(db_norm))
        state.ax.draw_artist(state.mesh)

        draw_title(
            state.ax_psd,
            state.psd_title,
            scan_duration,
            tune_step_hz,
            tune_step_fft,
            tune_rate_hz,
            tune_dwell_ms,
            config.sampling_rate,
            config.freq_resolution,
        )

        state.sm.set_clim(vmin=state.db_min, vmax=state.db_max)
        state.cbar.update_normal(state.sm)
        for ax in (state.cbar_ax.yaxis, state.ax_psd.yaxis):
            state.cbar_ax.draw_artist(ax)
            state.fig.canvas.blit(ax.axes.figure.bbox)
        for ln in state.top_n_lns:
            state.ax.draw_artist(ln)

        state.ax.draw_artist(state.ax.yaxis)
        for bmap in (
            state.ax_psd.bbox,
            state.ax.yaxis.axes.figure.bbox,
            state.ax.bbox,
            state.cbar_ax.bbox,
            state.fig.bbox,
        ):
            state.fig.canvas.blit(bmap)
        state.fig.canvas.flush_events()
        fig_path = None
        if config.savefig_path:
            fig_path = safe_savefig(config.savefig_path)

        if state.save_path:
            save_waterfall(
                state,
                config.save_time,
                scan_time,
                fig_path=fig_path,
            )


class WaterfallConfig:
    def __init__(
        self,
        engine,
        plot_snr,
        savefig_path,
        sampling_rate,
        fft_len,
        min_freq,
        max_freq,
        top_n,
        base_save_path,
        width,
        height,
        waterfall_height,
        waterfall_width,
        batch,
        rotate_secs,
        save_time,
    ):
        self.engine = engine
        self.plot_snr = plot_snr
        self.savefig_path = savefig_path
        self.snr_min = 0
        self.snr_max = 50
        self.waterfall_height = waterfall_height  # number of waterfall rows
        self.marker_distance = 0.1
        self.scale = 1e6
        self.fft_len = fft_len
        self.sampling_rate = sampling_rate
        self.psd_db_resolution = 90
        self.y_label_skip = 3
        self.top_n = top_n
        self.draw_rate = 1
        self.base_save_path = base_save_path
        self.width = width
        self.height = height
        self.batch = batch
        self.reclose_interval = 25
        self.min_freq = min_freq / self.scale
        self.max_freq = max_freq / self.scale
        self.freq_range = self.max_freq - self.min_freq
        self.freq_resolution = self.sampling_rate / self.scale / self.fft_len * 2
        self.waterfall_width = int(
            min(self.freq_range / self.freq_resolution, waterfall_width)
        )
        self.freq_resolution = self.freq_range / self.waterfall_width
        self.n_ticks = 20
        self.rotate_secs = rotate_secs
        self.save_time = save_time

    def __eq__(self, other):
        for attr in ("fft_len", "sampling_rate", "min_freq", "max_freq"):
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True


class WaterfallState:
    def __init__(self, config, save_path, peak_finder):
        self.db_min = -220
        self.db_max = -150
        self.detection_text = []
        self.scan_times = []
        self.scan_config_history = {}
        self.y_ticks = []
        self.y_labels = []
        self.previous_scan_config = None
        self.last_save_time = None
        self.counter = 0
        self.minor_tick_separator = None
        self.major_tick_separator = None
        self.cmap_psd = None
        self.cmap = None
        self.fig = None
        self.top_n_lns = None
        self.background = None
        self.mesh = None
        self.psd_title = None
        self.cbar_ax = None
        self.cbar = None
        self.psd_x_edges = None
        self.psd_y_edges = None
        self.min_psd_ln = None
        self.max_psd_ln = None
        self.mean_psd_ln = None
        self.current_psd_ln = None
        self.ax_psd = None
        self.ax = None
        self.save_path = save_path
        self.mesh_psd = None
        self.peak_finder = peak_finder
        self.last_plot = 0
        self.X, self.Y = meshgrid(
            config, 1, config.waterfall_height, config.waterfall_height
        )
        self.freq_bins = self.X[0]
        self.db_data = np.empty(self.X.shape)
        self.db_data.fill(np.nan)
        self.freq_data = np.empty(self.X.shape)
        self.freq_data.fill(np.nan)


def make_config(
    scan_configs,
    min_freq,
    max_freq,
    engine,
    plot_snr,
    savefig_path,
    top_n,
    base_save_path,
    width,
    height,
    waterfall_height,
    waterfall_width,
    batch,
    rotate_secs,
    save_time,
):
    sampling_rate = max([scan_config["sample_rate"] for scan_config in scan_configs])
    fft_len = max([scan_config["nfft"] for scan_config in scan_configs])
    if min_freq == 0:
        min_freq = min([scan_config["freq_start"] for scan_config in scan_configs])
    if max_freq == 0:
        max_freq = max([scan_config["freq_end"] for scan_config in scan_configs])

    config = WaterfallConfig(
        engine,
        plot_snr,
        savefig_path,
        sampling_rate,
        fft_len,
        min_freq,
        max_freq,
        top_n,
        base_save_path,
        width,
        height,
        waterfall_height,
        waterfall_width,
        batch,
        rotate_secs,
        save_time,
    )
    return config


def waterfall(
    min_freq,
    max_freq,
    plot_snr,
    top_n,
    base_save_path,
    save_time,
    peak_finder,
    engine,
    savefig_path,
    rotate_secs,
    width,
    height,
    waterfall_height,
    waterfall_width,
    batch,
    zmqr,
):
    style.use("fast")

    global need_reset_fig
    need_reset_fig = True
    global running
    running = True
    need_init = True
    need_reconfig = True

    def onresize(_event):  # nosemgrep
        global need_reset_fig
        need_reset_fig = True

    def sig_handler(_sig=None, _frame=None):
        global running
        running = False

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    logging.info("awaiting scanner startup")
    while not zmqr.healthy():
        time.sleep(0.1)

    logging.info("awaiting initial config from scanner(s)")
    scan_configs = None
    scan_df = None
    while zmqr.healthy() and running:
        scan_configs, scan_df = zmqr.read_buff(scan_fres=SCAN_FRES)
        if scan_df is not None:
            break
        time.sleep(0.1)

    if not scan_configs:
        return

    while zmqr.healthy() and running:
        if need_reconfig:
            config = make_config(
                scan_configs,
                min_freq,
                max_freq,
                engine,
                plot_snr,
                savefig_path,
                top_n,
                base_save_path,
                width,
                height,
                waterfall_height,
                waterfall_width,
                batch,
                rotate_secs,
                save_time,
            )
            logging.info(
                "scanning %fMHz to %fMHz at %fMsps with %u FFT points at %fMHz resolution",
                config.min_freq,
                config.max_freq,
                config.sampling_rate / 1e6,
                config.fft_len,
                config.freq_resolution,
            )

            state = WaterfallState(config, base_save_path, peak_finder)
            matplotlib.use(config.engine)
            results = [
                (scan_configs, frame_resample(scan_df, config.freq_resolution * 1e6))
            ]

            need_reconfig = False
            need_init = True
        if need_init:
            init_fig(config, state, onresize)
            need_init = False
            need_reset_fig = True
        if need_reset_fig:
            reset_fig(config, state)
            need_reset_fig = False
        while True:
            scan_configs, scan_df = zmqr.read_buff(
                scan_fres=config.freq_resolution * 1e6
            )
            if scan_df is None:
                break
            last_config = make_config(
                scan_configs,
                min_freq,
                max_freq,
                engine,
                plot_snr,
                savefig_path,
                top_n,
                base_save_path,
                width,
                height,
                waterfall_height,
                waterfall_width,
                batch,
                rotate_secs,
                save_time,
            )
            if last_config != config:
                logging.info("scanner config change detected")
                results = []
                need_reconfig = True
                break
            scan_df = scan_df[
                (scan_df.freq >= config.min_freq) & (scan_df.freq <= config.max_freq)
            ]
            if scan_df.empty:
                logging.info(
                    f"Scan is outside specified frequency range ({config.min_freq} to {config.max_freq})."
                )
                continue
            results.append((scan_configs, scan_df))
        if need_reconfig:
            continue
        if results:
            update_fig(config, state, results)
            if config.batch and state.counter % config.reclose_interval == 0:
                need_init = True
            results = []
        else:
            time.sleep(0.1)
    zmqr.stop()


class FlaskHandler:
    def __init__(self, savefig_path, port, refresh=5):
        self.savefig_path = savefig_path
        self.refresh = refresh
        self.app = Flask(__name__, static_folder=os.path.dirname(savefig_path))
        self.savefig_file = os.path.basename(self.savefig_path)
        self.app.add_url_rule("/", "", self.serve)
        self.app.add_url_rule(
            "/" + self.savefig_file, self.savefig_file, self.serve_waterfall
        )
        self.process = multiprocessing.Process(
            target=self.app.run,
            kwargs={"host": "0.0.0.0", "port": port},  # nosec
        )

    def start(self):
        self.process.start()

    def serve(self):
        if os.path.exists(self.savefig_path):
            return (
                '<html><head><meta http-equiv="refresh" content="%u"></head><body><img src="%s"></img></body></html>'
                % (self.refresh, self.savefig_file),
                200,
            )
        return (
            '<html><head><meta http-equiv="refresh" content="%u"></head><body>waterfall initializing, please wait or reload...</body></html>'
            % self.refresh,
            200,
        )

    def serve_waterfall(self):
        return current_app.send_static_file(self.savefig_file)


def main():
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")

    parser = argument_parser()
    args = parser.parse_args()
    detection_type = args.detection_type.lower()
    peak_finder = get_peak_finder(detection_type)

    if (detection_type and not args.n_detect) or (not detection_type and args.n_detect):
        raise ValueError("--detection_type and --n_detect must be set together")

    with tempfile.TemporaryDirectory() as tempdir:
        flask = None
        savefig_path = None
        engine = "GTK3Agg"
        batch = False

        if args.port:
            engine = "agg"
            batch = True
            savefig_path = os.path.join(tempdir, "waterfall.png")
            flask = FlaskHandler(savefig_path, args.port)
            flask.start()

        zmqr = ZmqReceiver(
            scanners=parse_scanners(args.scanners),
        )

        waterfall(
            args.min_freq,
            args.max_freq,
            args.plot_snr,
            args.n_detect,
            args.save_path,
            args.save_time,
            peak_finder,
            engine,
            savefig_path,
            args.rotate_secs,
            args.width,
            args.height,
            args.waterfall_height,
            args.waterfall_width,
            batch,
            zmqr,
        )


if __name__ == "__main__":
    main()



import argparse
import concurrent.futures
import json
import logging
import os
import subprocess
import threading
import time

import bjoern
import falcon
import jinja2
import numpy as np
import requests
import schedule

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import start_http_server

from gamutrf.sigwindows import calc_db
from gamutrf.sigwindows import choose_record_signal
from gamutrf.sigwindows import choose_recorders
from gamutrf.sigwindows import get_center
from gamutrf.sigwindows import parse_freq_excluded
from gamutrf.sigwindows import find_sig_windows
from gamutrf.sigwindows import ROLLING_FACTOR
from gamutrf.utils import rotate_file_n, SCAN_FRES
from gamutrf.zmqreceiver import ZmqReceiver, parse_scanners


MB = int(1.024e6)
PEAK_TRIGGER = int(os.environ.get("PEAK_TRIGGER", "0"))
PIN_TRIGGER = int(os.environ.get("PIN_TRIGGER", "17"))
if PEAK_TRIGGER == 1:
    import RPi.GPIO as GPIO

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(PIN_TRIGGER, GPIO.OUT)

PEAK_DBS = {}


def falcon_response(resp, text, status):
    resp.status = status
    resp.text = text
    resp.content_type = "text/html"


def ok_response(resp, text="ok!"):
    falcon_response(resp, text=text, status=falcon.HTTP_200)


def error_response(resp, text="error!"):
    falcon_response(resp, text=text, status=falcon.HTTP_500)


def load_template(name):
    path = os.path.join("templates", name)
    with open(os.path.abspath(path), "r", encoding="utf-8") as fp:
        return jinja2.Template(fp.read())


class ActiveRequests:
    def on_get(self, req, resp):
        all_jobs = schedule.get_jobs()
        ok_response(resp, f"{all_jobs}")


class ScannerForm:
    def on_get(self, req, resp):
        template = load_template("scanner_form.html")
        ok_response(resp, template.render(bins=PEAK_DBS))


class Result:
    def on_post(self, req, resp):
        # TODO validate input
        try:
            recorder = f'http://{req.media["worker"]}:8000/'
            signal_hz = int(int(req.media["frequency"]) * 1e6)
            record_bps = int(int(req.media["bandwidth"]) * MB)
            record_samples = int(record_bps * int(req.media["duration"]))
            recorder_args = f"record/{signal_hz}/{record_samples}/{record_bps}"
            timeout = int(req.media["duration"])
            response = None
            if int(req.media["repeat"]) == -1:
                schedule.every(timeout).seconds.do(
                    run_threaded,
                    record,
                    recorder=recorder,
                    recorder_args=recorder_args,
                    timeout=timeout,
                ).tag(f"{recorder}{recorder_args}-{timeout}")
                ok_response(resp)
            else:
                response = recorder_req(recorder, recorder_args, timeout)
                time.sleep(timeout)
                for _ in range(int(req.media["repeat"])):
                    response = recorder_req(recorder, recorder_args, timeout)
                    time.sleep(timeout)
                if response:
                    ok_response(resp)
                else:
                    ok_response(resp, f"Request {recorder} {recorder_args} failed.")
        except Exception as e:
            error_response(resp, f"{e}")


def record(recorder, recorder_args, timeout):
    recorder_req(recorder, recorder_args, timeout)


def run_threaded(job_func, recorder, recorder_args, timeout):
    job_thread = threading.Thread(
        target=job_func,
        args=(
            recorder,
            recorder_args,
            timeout,
        ),
    )
    job_thread.start()


def init_prom_vars():
    prom_vars = {
        "last_bin_freq_time": Gauge(
            "last_bin_freq_time",
            "epoch time last signal in each bin",
            labelnames=("bin_mhz",),
        ),
        "worker_record_request": Gauge(
            "worker_record_request",
            "record requests made to workers",
            labelnames=("worker",),
        ),
        "freq_power": Gauge(
            "freq_power", "bin frequencies and db over time", labelnames=("bin_freq",)
        ),
        "new_bins": Counter(
            "new_bins", "frequencies of new bins", labelnames=("bin_freq",)
        ),
        "old_bins": Counter(
            "old_bins", "frequencies of old bins", labelnames=("bin_freq",)
        ),
        "bin_freq_count": Counter(
            "bin_freq_count", "count of signals in each bin", labelnames=("bin_mhz",)
        ),
        "frame_counter": Counter("frame_counter", "number of frames processed"),
    }
    return prom_vars


def update_prom_vars(peak_dbs, new_bins, old_bins, prom_vars):
    freq_power = prom_vars["freq_power"]
    new_bins_prom = prom_vars["new_bins"]
    old_bins_prom = prom_vars["old_bins"]
    for freq in peak_dbs:
        freq_power.labels(bin_freq=freq).set(peak_dbs[freq])
    for nbin in new_bins:
        new_bins_prom.labels(bin_freq=nbin).inc()
    for obin in old_bins:
        old_bins_prom.labels(bin_freq=obin).inc()


def process_scan(args, scan_configs, prom_vars, df, lastbins):
    global PEAK_DBS
    df = calc_db(df, args.db_rolling_factor)
    freqdiffs = df.freq - df.freq.shift()
    mindiff = freqdiffs.min()
    maxdiff = freqdiffs.max()
    meandiff = freqdiffs.mean()
    logging.info(
        "new frame with %u samples, frequency sample differences min %f mean %f max %f",
        len(df),
        mindiff,
        meandiff,
        maxdiff,
    )
    if meandiff > mindiff * 2:
        logging.warning(
            "mean frequency diff larger than minimum - increase scanner sample rate"
        )
        logging.warning(df[freqdiffs > mindiff * 2])
    if args.fftlog:
        tmp_fftlog = os.path.join(
            os.path.dirname(args.fftlog), "." + os.path.basename(args.fftlog)
        )
        df.to_csv(tmp_fftlog, sep="\t", header=False, index=False)
        os.rename(tmp_fftlog, args.fftlog)
    monitor_bins = set()
    peak_dbs = {}
    bin_freq_count = prom_vars["bin_freq_count"]
    last_bin_freq_time = prom_vars["last_bin_freq_time"]
    freq_start_mhz = (
        min([scan_config["freq_start"] for scan_config in scan_configs]) / 1e6
    )
    signals = []
    for scan_config in scan_configs:
        scan_df = df[
            (df.freq >= scan_config["freq_start"] / 1e6)
            & (df.freq <= scan_config["freq_end"] / 1e6)
        ]
        logging.info(args.detection_type)
        if args.detection_type:
            signals.extend(
                find_sig_windows(
                    scan_df,
                    args.detection_type,
                )
            )
    min_samp_rate = (
        min([scan_config["sample_rate"] for scan_config in scan_configs]) / 1e6
    )
    df.loc[freqdiffs > min_samp_rate * 2, "db"] = np.nan

    if PEAK_TRIGGER == 1 and signals:
        led_sleep = 0.2
        GPIO.output(PIN_TRIGGER, GPIO.HIGH)
        time.sleep(led_sleep)
        GPIO.output(PIN_TRIGGER, GPIO.LOW)

    ts = df["ts"].max()
    for peak_freq, peak_db in signals:
        center_freq = get_center(
            peak_freq, freq_start_mhz, args.bin_mhz, args.record_bw_msps
        )
        logging.info(
            "detected peak at %f MHz %f dB, assigned bin frequency %f MHz",
            peak_freq,
            peak_db,
            center_freq,
        )
        bin_freq_count.labels(bin_mhz=center_freq).inc()
        last_bin_freq_time.labels(bin_mhz=ts).set(ts)
        monitor_bins.add(center_freq)
        peak_dbs[center_freq] = peak_db
    logging.info(
        "current bins %f to %f MHz: %s",
        df["freq"].min(),
        df["freq"].max(),
        sorted(peak_dbs.items()),
    )
    PEAK_DBS = sorted(peak_dbs.items())
    new_bins = monitor_bins - lastbins
    if new_bins:
        logging.info("new bins: %s", sorted(new_bins))
    old_bins = lastbins - monitor_bins
    if old_bins:
        logging.info("old bins: %s", sorted(old_bins))
    update_prom_vars(peak_dbs, new_bins, old_bins, prom_vars)
    return monitor_bins


def recorder_req(recorder, recorder_args, timeout):
    url = f"{recorder}/v1/{recorder_args}"
    try:
        req = requests.get(url, timeout=timeout)
        logging.debug(str(req))
        return req
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as err:
        logging.debug(str(err))
        return None


def get_freq_exclusions(args):
    recorder_freq_exclusions = {}
    for recorder in args.recorder:
        req = recorder_req(recorder, "info", args.record_secs)
        if req is None or req.status_code != 200:
            continue
        excluded = json.loads(req.text).get("freq_excluded", None)
        if excluded is None:
            continue
        recorder_freq_exclusions[recorder] = parse_freq_excluded(excluded)
    return recorder_freq_exclusions


def call_record_signals(args, lastbins_history, prom_vars):
    if lastbins_history:
        signals = []
        for bins in lastbins_history:
            signals.extend(list(bins))
        recorder_freq_exclusions = get_freq_exclusions(args)
        recorder_count = len(recorder_freq_exclusions)
        record_signals = choose_record_signal(
            signals, recorder_count * args.max_recorder_signals
        )
        for signal, recorder in choose_recorders(
            record_signals, recorder_freq_exclusions, args.max_recorder_signals
        ):
            signal_hz = int(signal * 1e6)
            record_bps = int(args.record_bw_msps * MB)
            record_samples = int(record_bps * args.record_secs)
            recorder_args = f"record/{signal_hz}/{record_samples}/{record_bps}"
            resp = recorder_req(recorder, recorder_args, args.record_secs)
            if resp:
                worker_record_request = prom_vars["worker_record_request"]
                worker_record_request.labels(worker=recorder).set(signal_hz)


def zstd_file(uncompressed_file):
    subprocess.check_call(["/usr/bin/zstd", "--force", "--rm", uncompressed_file])


class ScanProcessor:
    def __init__(self, args, prom_vars):
        self.args = args
        self.prom_vars = prom_vars
        self.lastbins_history = []
        self.lastbins = set()
        self.frame_counter = self.prom_vars["frame_counter"]


def process_scans(args, prom_vars, executor, zmqr):
    sp = ScanProcessor(args, prom_vars)

    while True:
        if os.path.exists(args.log):
            logging.info(f"{args.log} exists, will append first")
            mode = "a"
        else:
            logging.info(f"opening {args.log}")
            mode = "w"
        deadline = int(time.time()) + args.rotatesecs
        with open(args.log, mode=mode, encoding="utf-8") as l:
            while True:
                now = int(time.time())
                if not zmqr.healthy():
                    return
                if now > deadline:
                    break
                scan_configs, frame_df = zmqr.read_buff(l, scan_fres=SCAN_FRES)
                if frame_df is None:
                    schedule.run_pending()
                    sleep_time = 1
                    time.sleep(sleep_time)
                    continue
                sp.frame_counter.inc()
                logging.info(
                    "frame with sweep_start %us ago",
                    now - frame_df["sweep_start"].min(),
                )
                new_lastbins = process_scan(
                    args,
                    scan_configs,
                    prom_vars,
                    frame_df,
                    sp.lastbins,
                )
                if new_lastbins is not None:
                    sp.lastbins = new_lastbins
                    if sp.lastbins:
                        sp.lastbins_history = [sp.lastbins] + sp.lastbins_history
                        sp.lastbins_history = sp.lastbins_history[: args.history]
                call_record_signals(args, sp.lastbins_history, prom_vars)
        rotate_file_n(".".join((args.log, "zst")), args.nlog, require_initial=False)
        new_log = ".".join((args.log, "1"))
        os.rename(args.log, new_log)
        executor.submit(zstd_file, new_log)


def argument_parser():
    parser = argparse.ArgumentParser(
        description="watch a scan UDP stream and find signals"
    )
    parser.add_argument(
        "--log", default="scan.log", type=str, help="base path for scan logging"
    )
    parser.add_argument(
        "--fftlog",
        default="",
        type=str,
        help="if defined, path to log last complete FFT frame",
    )
    parser.add_argument(
        "--rotatesecs",
        default=3600,
        type=int,
        help="rotate scan log after this many seconds",
    )
    parser.add_argument(
        "--nlog", default=10, type=int, help="keep only this many scan.logs"
    )
    parser.add_argument(
        "--bin_mhz", default=20, type=int, help="monitoring bin width in MHz"
    )
    parser.add_argument(
        "--history",
        default=5,
        type=int,
        help="number of frames of signal history to keep",
    )
    parser.add_argument(
        "--recorder",
        action="append",
        default=[],
        help="SDR recorder base URLs (e.g. http://host:port/, multiples can be specified)",
    )
    parser.add_argument(
        "--record_bw_msps",
        default=20,
        type=int,
        help="record bandwidth in n * {MB} samples per second",
    )
    parser.add_argument(
        "--record_secs", default=10, type=int, help="record time duration in seconds"
    )
    parser.add_argument(
        "--promport",
        dest="promport",
        type=int,
        default=9000,
        help="Prometheus client port",
    )
    parser.add_argument(
        "--port",
        dest="port",
        type=int,
        default=80,
        help="control webserver port",
    )
    parser.add_argument(
        "--scanners",
        dest="scanners",
        type=str,
        default="127.0.0.1:8001",
        help="Connect to gamutRF scanners at these addresses",
    )
    parser.add_argument(
        "--max_recorder_signals",
        dest="max_recorder_signals",
        type=int,
        default=1,
        help="Max number of recordings per worker to request",
    )
    parser.add_argument(
        "--buff_path",
        dest="buff_path",
        type=str,
        default="/dev/shm",  # nosec
        help="Path for FFT buffer file",
    )
    parser.add_argument(
        "--db_rolling_factor",
        dest="db_rolling_factor",
        type=float,
        default=0,
        help="Divisor for rolling dB average (or 0 to disable)",
    )
    parser.add_argument(
        "--detection_type",
        default="narrowband",
        type=str,
        help="Detection type to plot (wideband, narrowband).",
    )
    return parser


def main():
    parser = argument_parser()
    args = parser.parse_args()
    scanners = parse_scanners(args.scanners)

    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")
    prom_vars = init_prom_vars()
    app = falcon.App()
    scanner_form = ScannerForm()
    result = Result()
    active_requests = ActiveRequests()
    app.add_route("/", scanner_form)
    app.add_route("/result", result)
    app.add_route("/requests", active_requests)
    start_http_server(args.promport)

    with concurrent.futures.ProcessPoolExecutor(2) as executor:
        zmqr = ZmqReceiver(
            scanners=scanners,
            buff_path=args.buff_path,
        )
        x = threading.Thread(
            target=process_scans,
            args=(
                args,
                prom_vars,
                executor,
                zmqr,
            ),
        )
        x.start()
        try:
            bjoern.run(app, "0.0.0.0", args.port)  # nosec
        except KeyboardInterrupt:
            logging.info("interrupt!")
            zmqr.stop()
            executor.shutdown()
            x.join()



import numpy as np
from scipy.signal import find_peaks


class PeakFinderBase:
    name = "base"

    def find_peaks(self, db_data, height=None):
        raise NotImplementedError


class PeakFinderNarrow:
    name = "narrowband"

    def find_peaks(self, db_data, height=None):
        if height is None:
            height = np.nanmean(db_data)
        peaks, properties = find_peaks(
            db_data,
            height=height,
            width=(1, 10),
            prominence=10,
            rel_height=0.7,
            wlen=120,
        )
        return peaks, properties


class PeakFinderWide:
    name = "wideband"

    def find_peaks(self, db_data, height=None):
        if height is None:
            height = np.nanmean(db_data) + 1
        peaks, properties = find_peaks(
            db_data,
            height=height,
            width=10,
            prominence=(0, 20),
            rel_height=0.7,
            wlen=120,
        )
        return peaks, properties


peak_finders = {
    "wideband": PeakFinderWide,
    "narrowband": PeakFinderNarrow,
    "wb": PeakFinderWide,
    "nb": PeakFinderNarrow,
}


def get_peak_finder(name):
    if not name:
        return None
    try:
        return peak_finders[name]()
    except KeyError:
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import os
import sys
from pathlib import Path
import cv2
import numpy as np

try:
    from gnuradio import gr  # pytype: disable=import-error
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)


DELIM = "\n\n"


class yolo_bbox(gr.sync_block):
    def __init__(
        self,
        output_dir,
        confidence_threshold=0.5,
        nms_threshold=0.5,
    ):
        self.output_dir = output_dir
        self.confidence_threshold = confidence_threshold
        self.nms_threshold = nms_threshold
        self.yaml_buffer = ""

        gr.sync_block.__init__(
            self,
            name="yolo_bbox",
            in_sig=[np.ubyte],
            out_sig=None,
        )

    def work(self, input_items, output_items):
        n = 0
        for input_item in input_items:
            raw_input_item = input_item.tobytes().decode("utf8")
            n += len(raw_input_item)
            self.yaml_buffer += raw_input_item
        while True:
            delim_pos = self.yaml_buffer.find(DELIM)
            if delim_pos == -1:
                break
            raw_item = self.yaml_buffer[:delim_pos]
            item = json.loads(raw_item)
            self.yaml_buffer = self.yaml_buffer[delim_pos + len(DELIM) :]
            self.process_item(item)
        return n

    def draw_bounding_box(self, img, name, confidence, x, y, x_plus_w, y_plus_h):
        label = f"{name}: {confidence}"
        color = (255, 255, 255)
        cv2.rectangle(img, (x, y), (x_plus_w, y_plus_h), color, 2)
        cv2.putText(
            img, label, (x - 10, y - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color, 2
        )

    def process_item(self, item):
        predictions = item.get("predictions", None)
        if not predictions:
            return

        boxes = []
        scores = []
        detections = []

        try:
            for name, prediction_data in predictions.items():
                for prediction in prediction_data:
                    conf = prediction["conf"]
                    if conf < self.confidence_threshold:
                        continue
                    xywh = prediction["xywh"]
                    box = [
                        xywh[0] - (0.5 * xywh[2]),
                        xywh[1] - (0.5 * xywh[3]),
                        xywh[2],
                        xywh[3],
                    ]
                    detections.append({"box": box, "score": conf, "name": name})
                    boxes.append(box)
                    scores.append(conf)
        except TypeError as e:
            print(f"invalid predictions from torchserve: {e}, {predictions}")
            return

        if not detections:
            return

        original_image = cv2.imread(item["image_path"])
        result_boxes = cv2.dnn.NMSBoxes(
            boxes, scores, self.confidence_threshold, self.nms_threshold, 0.5, 200
        )

        # TODO: output to ZMQ
        for detection in detections:
            self.draw_bounding_box(
                original_image,
                detection["name"],
                detection["score"],
                round(box[0]),
                round(box[1]),
                round(box[0] + box[2]),
                round(box[1] + box[3]),
            )

        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        filename = str(
            Path(
                self.output_dir,
                "_".join(["prediction", os.path.basename(item["image_path"])]),
            )
        )
        cv2.imwrite(filename, original_image)
